"""Reachy top-level module.

Exposes:

* :py:mod:`~reachy.__version__` number
* :py:class:`~Reachy` class
* :py:mod:`~reachy.parts` submodule
"""

from ._version import __version__  # noqa: F401
from .reachy import Reachy  # noqa: F401
